# Les performances en Python

par

<!-- .slide: data-background="static/background.jpg" -->

Julien Palard <julien@palard.fr>

https://mdk.fr


# Bien choisir sa structure de donnée

C'est bien choisir l'algorihtme qu'on va utiliser.


## Comparaison asymptotique

Les notations les plus utilisées :

```text
O(1)        Constant
O(log n)    Logarithmique
O(n)        Linéaire
O(n log n)  Parfois appelée « linéarithmique »
O(n²)       Quadratique
O(nᶜ)       Polynomiale
O(cⁿ)       Exponentielle
O(n!)       Factorielle
```

notes:

Il faut les grapher pour s'en rendre compte : cf. examples/big.o.py


## Comparaison asymptotique

Exemples.


## O(1)

```python
def get_item(a_list: list, an_idx):
    return a_list[an_idx]
```

ou

```python
def is_in(a_set: set, a_value):
    return a_value in a_set
```

notes:

Peu importe la taille de la liste, accéder à un élément prend le même temps.


## O(log n)

Attention c'est toujours en base deux.

Exemple typique : chercher dans un annuaire.

notes:

Un annuaire deux fois plus gros ne vous demandera pas deux fois plus
de temps mais peut-être une opération de plus.


## O(log n)

```python
#!sed -n '/def index/,/raise ValueError/p' examples/find_in_list.py
```

## O(n)

```python
#!sed -n '/def dumb_index/,/raise ValueError/p' examples/find_in_list.py
```

## O(n log n)

C'est `n` fois un `log n`, par exemple rayer `n` personnes dans un annuaire.

Typique d'algorithmes de tris.


## Les mesures de complexité

- De temps (CPU consommé).
- D'espace (mémoire  consommée).
- Dans le meilleur des cas.
- Dans le pire des cas.
- Dans le cas moyen.
- Amorti.
- ...


## Les mesures de complexité

Il n'est pas forcément nécessaire d'apprendre par cœur toutes les complexités de chaque opération.

Pas toute suite.


## Les bases

Mais retenir par cœur la complexité de quelques structures
élémentaires permet d'éviter les « erreurs de débutants ».


## Rappel des unités de temps

- 1 milliseconde (1 ms) c'est un millième de seconde.
- 1 microseconde (1 μs) c'est un millionième de seconde.
- 1 nanoseconde (1 ns) c'est un milliardième de seconde.


## Rappel des unités de temps

- milli c'est `10 ** -3`, c'est `0.001`.
- micro c'est `10 ** -6`, c'est `0.000_001`.
- nano c'est `10 ** -9`, c'est `0.000_000_001`.


## Le cas typique

```shell
#!cache python -m pyperf timeit --setup 'container = list(range(10_000_000))' '10_000_001 in container'

#!cache python -m pyperf timeit --setup 'container = set(range(10_000_000))' '10_000_001 in container'
```

Pourquoi une si grande différence !?


notes:

C'est l'heure du live coding !


# À vous !

Simulons un tas de sable.

Ne vous souciez pas des perfs, on s'en occupera.

Vous avez 10mn.

notes:

voir sandpile.py


# Les outils

## Les outils en ligne de commande

`time`, un outil POSIX, mais aussi une fonction native de bash :

```shell
#!cache time -p python examples/sandpile.py 16
```

Mais `time` ne teste qu'une fois.

notes:

real    0m0.719s  # C'est le temps « sur le mur »
user    0m0.521s  # Temps CPU passé « dans Python »
sys     0m0.195s  # Temps CPU passé dans le kernel


## Hyperfine

`hyperfine` teste plusieurs fois :

```shell
#!cache hyperfine "python examples/sandpile.py 1000"
```


## Petite parenthèse

Mais attention, démarrer un processus Python n'est pas gratuit :

```shell
#!cache hyperfine "python -c pass"
```

notes:

N'essayez pas de retenir les chiffres, retenez les faits.


## Petite parenthèse

Et puis il peut dépendre de la version de Python, des options de compilation, ... :

```shell
$ hyperfine "~/.local/bin/python3.10 -c pass"  # Avec pydebug
Benchmark 1: ~/.local/bin/python3.10 -c pass
  Time (mean ± σ):      37.6 ms ±   0.6 ms

$ hyperfine "/usr/bin/python3.10 -c pass"  # Sans pydebug
Benchmark 1: /usr/bin/python3.10 -c pass
  Time (mean ± σ):      19.1 ms ±   0.8 ms
```

notes:

Leur parler de `--enable-optimizations` (PGO).


## timeit

Timeit c'est dans la stdlib de Python, ça s'utilise en ligne de commande ou depuis Python.

```shell
#!cache python -m timeit -s 'import sandpile' 'sandpile.main(1000, show=False)'
```

Mais il ne vous donne que le meilleur temps, ce qui n'est pas toujours représentatif.


## pyperf

Il n'est pas dans la stdlib, mais il replace avantageusement `timeit` :

```shell
#!cache python -m pyperf timeit --setup 'import sandpile' 'sandpile.main(1000, show=False)'
```

notes:

Avec hyperfine on teste combien de temps ça prend à Python **de
démarrer** puis d'exécuter `pass`, ici on teste combien de temps ça
prend d'exécuter `pass`.


## pyperf

Digression, `pyperf` peut aussi remplacer `hyperfine` :

```shell
#!cache python -m pyperf command python examples/sandpile.py 1000
```

notes:

Avec hyperfine on teste combien de temps ça prend à Python **de
démarrer** puis d'exécuter `pass`, ici on teste combien de temps ça
prend d'exécuter `pass`.


## Les outils — À vous !

Effectuez quelques mesures sur votre implémentation.

Tentez d'en déterminer la complexité en fonction du nombre de grains.


# Profilage

`time`, `timeit`, `hyperfine`, `pyperf` c'est bien pour mesurer, comparer.

`cProfile` peut nous aider à trouver la fonction coupable.


## cProfile, exemple

```python
#!sed -n '/def fib/,/return approx/p' examples/phi1.py
```


## cProfile, exemple

Testons :

```python
#!sed -n '/import sys/,$p' examples/phi1.py
```

```text
#!cache time -p python examples/phi1.py 10
```


C'est déjà lent, et pour `20` c'est interminable...


## cProfile, exemple

Sortons cProfile :

```shell
$ python -m cProfile --sort cumulative phi1.py 10
...
#!cache python -m cProfile --sort cumulative examples/phi1.py 10 | sed -n '/fib\|function calls/{s/ \+/ /g;s/^ *//;p}'
...
```

C'est donc `fib` la coupable :
- C'est ~100% du temps (`cumtime`).
- C'est ~100% des appels de fonctions.


## cProfile, exemple

Cachons les résultats de `fib` :
```python
#!sed -n '/import cache/,/return fib/p' examples/phi2.py
```

## cProfile, exemple

Et on repasse dans cProfile !

```shell
$ python -m cProfile --sort cumulative phi2.py 10
...
#!cache python -m cProfile --sort cumulative examples/phi2.py 10 | sed -n '/fib\|function calls/{s/ \+/ /g;s/^ *//;p}'
...
```

C'est mieux !

## cProfile, exemple

On essaye d'aller plus loin ?

```shell
#!cache python -m cProfile --sort cumulative examples/phi2.py 2000 | head -n 3 | sed 's/^ *//g;s/seconds/s/g'
```

Ça tient, mais peut-on faire mieux ?


## cProfile, exemple

Divisons par 10 le nombre d'appels, on réduira mécaniquement par 10 le
temps d'exécution ?

```python
#!sed -n '/def approx_phi_up_to/,/return step1/p' examples/phi3.py
```

## cProfile, exemple

```shell
#!cache python -m cProfile --sort cumulative examples/phi3.py 2000 | head -n 3 | sed 's/^ *//g;s/seconds/s/g'
```


## cProfile, exemple

En cachant `approx_phi` ?

```python
#!sed -n '10,/return step1/p' examples/phi4.py
```

notes:

Notez l'astuce pour que le `step2` d'un
tour soit le `step1` du suivant...


## cProfile, exemple

```shell
$ python -m cProfile --sort cumulative examples/phi4.py 2000
```

`RecursionError` !? En effet, en avançant par si grands pas, le cache
de `fib` n'est pas chaud, et il peut vite devoir descendre
profondément en récursion...


## cProfile, exemple

Il est temps de sortir une implémentation de `fib` plus robuste, basée
sur l'algorithme « matrix exponentiation » :

```python
#!sed -n '/def fib/,/return fib/p' examples/phi5.py
```


## cProfile, exemple

```text
#!cache python -m cProfile --sort cumulative examples/phi5.py 2000 | head -n 3 | sed 's/^ *//g;s/seconds/s/g'
```

notes:

Mieux.


## Snakeviz

```shell
$ python -m pip install snakeviz
$ python -m cProfile -o phi5.prof phi5.py 2000
$ python -m snakeviz phi5.prof
```

#!if [ ! -f .cache/phi5.prof ]; then python -m cProfile -o .cache/phi5.prof examples/phi5.py 2000 >/dev/null 2>&1; fi
#!if [ ! -f output/phi5-snakeviz.png ]; then python -m snakeviz -s .cache/phi5.prof & TOKILL=$!; sleep 1; cutycapt --min-width=1024 --delay=500 --url=http://127.0.0.1:8080/snakeviz/%2Ftmp%2Fphi5.prof --out=output/phi5-snakeviz.png ; kill $TOKILL; fi


## Snakeviz

![](phi5-snakeviz.png)


## Scalene

```shell
$ python -m pip install scalene
$ scalene phi5.py 100000
```

#!if [ ! -f output/phi5.html ]; then ( cd examples; scalene phi5.py 100000 --html --outfile ../output/phi5.html --cli >&2 ); fi
#!if [ ! -f output/phi5-scalene.png ]; then cutycapt --min-width=1024 --delay=100 --url=file://$(pwd)/output/phi5.html --out=output/phi5-scalene.png; fi


## Scalene

![](phi5-scalene.png)


## Aussi

- https://github.com/gaogaotiantian/viztracer
- https://github.com/joerick/pyinstrument
- https://github.com/benfred/py-spy
- https://github.com/pyutils/line_profiler
- https://github.com/sumerc/yappi
- https://github.com/vmprof/vmprof-python
- https://github.com/bloomberg/memray


## Profilage — À vous !

Profilez votre implémentation et tentez quelques améliorations.


# Cython

Cython est un dialecte de Python transpilable en C.


## Cython démo

```python
#!cat examples/collatz_length.py
```


## Cython démo

```shell
$ python -m pyperf timeit \
> -s 'from collatz_length import collatz_length'
> 'collatz_length(837799)'
#!cache python -m pyperf timeit --fast --setup 'from examples.collatz_length import collatz_length' 'collatz_length(837799)'
```

```shell
$ cythonize --inplace collatz_length_cython.py
#!if ! [ -f examples/collatz_length_cython.*.so ] ; then cythonize --inplace examples/collatz_length_cython.py; fi
```

```shell
$ python -m pyperf timeit \
> -s 'from collatz_length import collatz_length'
> 'collatz_length(837799)'
#!cache python -m pyperf timeit --fast -s 'from examples.collatz_length import collatz_length' 'collatz_length(837799)'  # faster
#!# Beware, the cythonized use `-s` while the non cythonized uses `--setup` just to have two cache buckets :D
```


## Cython annotate

```shell
$ cython -a collatz_length.py
#!if ! [ -f examples/collatz_length.html ] ; then cython -a examples/collatz_length.py; fi
#!if ! [ -f output/collatz_length.png ] ; then cutycapt --min-width=1024 --delay=500  --url=file://$(pwd)/examples/collatz_length.html --out=output/collatz_length.png; fi
```

![](collatz_length.png)


## Cython annotated

```python
#!cat examples/collatz_length_annotated.py
```

```shell
$ cythonize --inplace collatz_length_annotated.py
```

#!if ! [ -f examples/collatz_length_annotated.*.so ] ; then cythonize --inplace examples/collatz_length_annotated.py; fi

```shell
$ python -m pyperf timeit \
> -s 'from collatz_length_annotated import collatz_length'
> 'collatz_length(837799)'
#!cache python -m pyperf timeit --fast -s 'from examples.collatz_length_annotated import collatz_length' 'collatz_length(837799)'
```


## Cython annotate again

```shell
$ cython -a examples/collatz_length_annotated.py
#!if ! [ -f examples/collatz_length_annotated.html ] ; then cython -a examples/collatz_length_annotated.py; fi
#!if ! [ -f output/collatz_length_annotated.png ] ; then cutycapt --min-width=1024 --delay=500  --url=file://$(pwd)/examples/collatz_length_annotated.html --out=output/collatz_length_annotated.png; fi
```

![](collatz_length_annotated.png)


## Cython — À vous !


# Numba

Numba est un `JIT` : « Just In Time compiler ».

```python
#!cat examples/collatz_length_numba.py
```

## Numba démo

```shell
$ python -m pyperf timeit \
> -s 'from examples.collatz_length_numba import collatz_length' \
> 'collatz_length(837799)'
#!cache python -m pyperf timeit --fast -s 'from examples.collatz_length_numba import collatz_length' 'collatz_length(837799)'
```

## numba — À vous !


# mypyc

mypyc est un compilateur qui s'appuie sur les annotationes de type mypy :

```python
#!cat examples/collatz_length_mypy.py
```

## mypyc demo

```shell
$ mypyc examples/collatz_length_mypy.py
#!if ! [ -f collatz_length_mypy.*.so ] ; then mypyc examples/collatz_length_mypy.py; fi
```

```shell
$ python -m pyperf timeit \
> -s 'from collatz_length_mypy import collatz_length' \
> 'collatz_length(837799)'
#!cache python -m pyperf timeit --fast -s 'from collatz_length_mypy import collatz_length' 'collatz_length(837799)'
```

## mypyc — À vous !


# Pythran

pythran est un compilateur pour du code scientifique :

```python
#!cat examples/collatz_length_pythran.py
```

## Pythran demo

```shell
$ pythran examples/collatz_length_pythran.py
#!if ! [ -f collatz_length_pythran.*.so ]; then pythran examples/collatz_length_pythran.py; fi
```

```shell
$ python -m pyperf timeit \
> -s 'from collatz_length_pythran import collatz_length' \
> 'collatz_length(837799)'
#!cache python -m pyperf timeit --fast -s 'from collatz_length_pythran import collatz_length' 'collatz_length(837799)'
```

## pythran — À vous !


# Nuitka

Aussi un compilateur, aussi utilisable pour distribuer une application.

```shell
$ python -m nuitka --module collatz_length_nuitka.py
#!if ! [ -f collatz_length_nuitka.*.so ]; then python -m nuitka --module examples/collatz_length_nuitka.py >/dev/null; fi
```

```shell
$ python -m pyperf timeit \
> -s 'from collatz_length_nuitka import collatz_length' \
> 'collatz_length(837799)'
#!cache python -m pyperf timeit --fast -s 'from collatz_length_nuitka import collatz_length' 'collatz_length(837799)'
```


# Et d'autres…

https://github.com/pfalcon/awesome-python-compilers


# Hand crafted C

```c
#!sed -n '/int collatz_length/,/^$/p' examples/my_collatz_length.c
```

Mais comment l'utiliser ?

## Hand crafted C

### Avec Cython

```cpython
#!cat examples/collatz_length_cython_to_c.pyx
```

```shell
$ cythonize -i examples/collatz_length_cython_to_c.pyx
#!if ! [ -f examples/collatz_length_cython_to_c.*.so ] ; then cythonize -i examples/collatz_length_cython_to_c.pyx; fi
```

```shell
$ python -m pyperf timeit \
> -s 'from collatz_length_cython_to_c import collatz_length' \
> 'collatz_length(837799)'
#!cache python -m pyperf timeit --fast -s 'from examples.collatz_length_cython_to_c import collatz_length' 'collatz_length(837799)'
```
